from pathlib import (
    Path,
)

from edu_rdm_integration.export_data.base.requests import (
    RegionalDataMartEntityRequest,
)
from edu_rdm_integration.management.general import (
    BaseDatamartClientCommand,
)


class Command(BaseDatamartClientCommand):

    help = 'Команда для загрузки csv-файла в РВД с использованием uploader-client.'  # noqa A003

    def add_arguments(self, parser):
        """Добавление параметров."""
        super().add_arguments(parser)

        parser.add_argument(
            '--table_name',
            type=str,
            required=True,
            help='название таблицы внутри витрины данных',
        )
        parser.add_argument(
            '--file_path',
            type=str,
            required=True,
            help='путь до csv-файла с данными',
        )

    def _get_request(self, **options):
        """Возвращает запрос для отправки в РВД."""
        request = RegionalDataMartEntityRequest(
            datamart_name=options['datamart_mnemonic'],
            table_name=options['table_name'],
            method='POST',
            operation='upload',
            parameters={},
            headers={
                'Content-Type': 'text/csv',
            },
            files=[],
            data=Path(options['file_path']).open('rb').read(),
        )

        return request



from educommon.integration_entities.enums import (
    EntityLogOperation,
)


REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA = 'regional_data_mart_integration_collecting_data'
REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA = 'regional_data_mart_integration_exporting_data'

# Формат даты. Используется для выгрузки
DATE_FORMAT = '%Y-%m-%d'

# Формат даты/времени. Для выгрузки не используется (в выгрузке ISO формат)
DATETIME_FORMAT = '%d.%m.%Y %H:%M:%S'

# Формат даты/времени. Для выгрузки
EXPORT_DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'

LOGS_DELIMITER = '    '

HASH_ALGORITHM = 'md_gost12_512'

BATCH_SIZE = 5000

ACADEMIC_YEAR = {
    'start_day': 1,
    'start_month': 9,
    'end_day': 31,
    'end_month': 8,
}

TASK_QUEUE_NAME = 'RDM'


# Кортеж операций для обновления данных
UPDATED_OPERATIONS = (EntityLogOperation.CREATE, EntityLogOperation.UPDATE)
ALL_OPERATIONS = UPDATED_OPERATIONS + (EntityLogOperation.DELETE, )



from typing import (
    Sequence,
    TypeVar,
    Union,
)

from django.db.models import (
    Expression,
    Model,
)


# Тип, обозначающий любую модель. При указании в качестве type annotation
# можно указать, что аргументом может быть любая модель или тип модели (через
# Type[MODEL_TYPE_VAR]), а сама функция возвращает инстанс этой
# конкретной модели
MODEL_TYPE_VAR = TypeVar('MODEL_TYPE_VAR', bound=Model)

# Аннотация типов для id записи в БД
RECORD_IDS = Union[
    tuple[int, ...],
    tuple[str, ...],
    list[int],
    list[str],
    Sequence[int],
    Sequence[str],
    Expression,
]



from educommon import (
    ioc,
)

from edu_rdm_integration.registry.actions import (
    EntitySelectPack,
    TransferredEntityPack,
)


def register_actions():
    """Регистрирует паки и экшны."""
    ioc.get('main_controller').packs.extend((
        TransferredEntityPack(),
        EntitySelectPack(),
    ))



from edu_rdm_integration.export_data.base.requests import (
    RegionalDataMartStatusRequest,
)
from edu_rdm_integration.management.general import (
    BaseDatamartClientCommand,
)


class Command(BaseDatamartClientCommand):

    help = 'Команда для получения статуса загрузки файла в РВД с использованием uploader-client.'  # noqa A003

    def add_arguments(self, parser):
        """Добавление параметров."""
        super().add_arguments(parser)

        parser.add_argument(
            '--request_id',
            type=str,
            required=True,
            help='requestID полученный в результате выполнения команды на загрузку или удаление данных',
        )

    def _get_request(self, **options):
        """Возвращает запрос для отправки в РВД."""
        request = RegionalDataMartStatusRequest(
            request_id=options['request_id'],
            method='GET',
            parameters={},
            headers={
                'Content-Type': 'application/json',
            },
        )

        return request



from m3.db import (
    BaseEnumerate,
)


class RequestResultStatus(BaseEnumerate):
    """Результат выполнения запроса."""

    SUCCESS = 1
    ERROR = 2

    values = {
        SUCCESS: 'Успех',
        ERROR: 'Ошибка',
    }



import os
from datetime import (
    date,
    datetime,
    time,
    timedelta,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
)

from django.conf import (
    settings,
)
from django.db.models import (
    DateField,
    DateTimeField,
    Expression,
    F,
)
from django.db.models.expressions import (
    Func,
)
from django.db.models.functions import (
    Cast,
    Now,
)
from educommon.integration_entities.enums import (
    EntityLogOperation,
)

from edu_rdm_integration.apps import (
    EduRDMIntegrationConfig,
)
from edu_rdm_integration.consts import (
    ACADEMIC_YEAR,
)


if TYPE_CHECKING:
    from typing import (
        NamedTuple,
    )

    from django.db.models import (
        Model,
    )

    from edu_rdm_integration.collect_and_export_data.models import (
        EduRdmCollectDataCommandProgress,
        EduRdmExportDataCommandProgress,
    )
    from edu_rdm_integration.helpers import (
        Graph,
    )
    from edu_rdm_integration.models import (
        BaseEntityModel,
    )


def get_exporting_data_stage_attachment_path(instance, filename):
    """Возвращает путь до файла-вложения в этап выгрузки данных сущности.

    Args:
        instance: объект ExportingDataStage
        filename: имя загружаемого файла

    Returns:
        Строковое представление пути
    """
    datetime_now = datetime.now()

    return os.path.join(
        settings.UPLOADS,
        EduRDMIntegrationConfig.label,
        datetime_now.strftime('%Y/%m/%d'),
        instance.exporting_data_sub_stage.__class__.__name__.lower(),
        str(instance.exporting_data_sub_stage_id),
        str(instance.operation),
        filename,
    )


def update_fields(
    entity: 'BaseEntityModel',
    field_values: Dict[str, Any],
    mapping: Dict[str, str]
) -> None:
    """Обновление значений полей сущности по измененным полям модели.

    :param entity: Выгружаемая сущность
    :param field_values: Словарь с измененными данными модели
    :param mapping: Словарь маппинга полей модели к полям сущности
    """
    for model_field, entity_field in mapping.items():
        if model_field in field_values:
            setattr(entity, entity_field, field_values[model_field])


def set_entity_field_by_model_object(
    entity: 'BaseEntityModel',
    model_object: Union['Model', 'NamedTuple'],
    mapping: Dict[str, str]
) -> None:
    """Обновление значений полей сущности по измененным полям модели.

    Args:
        entity: Выгружаемая сущность;
        model_object: Объект модели с измененными полями;
        mapping: Словарь маппинга полей модели к полям сущности
    """
    for model_field, entity_field in mapping.items():
        if hasattr(model_object, model_field):
            setattr(entity, entity_field, getattr(model_object, model_field))


def get_isoformat_timezone():
    """Возвращает временную зонну в ISO представлении."""
    return datetime.now().astimezone().isoformat()[-6:]


def split_by_academic_years(start_date: datetime, end_date: datetime) -> List[tuple]:
    """Разбивает исходный интервал из команды по учебным годам."""
    academic_year_end_date = datetime(
        start_date.year, ACADEMIC_YEAR['end_month'], ACADEMIC_YEAR['end_day'],

    )
    if academic_year_end_date <= start_date:
        academic_year_end_date = datetime(
            academic_year_end_date.year + 1,
            ACADEMIC_YEAR['end_month'],
            ACADEMIC_YEAR['end_day'],
            time.max.hour,
            time.max.minute,
            time.max.second,
        )

    splitted_interval = []
    splitted_interval.append(start_date)

    while start_date < academic_year_end_date < end_date:
        splitted_interval.append(academic_year_end_date)

        academic_year_end_date = datetime(
            academic_year_end_date.year + 1,
            ACADEMIC_YEAR['end_month'],
            ACADEMIC_YEAR['end_day'],
            time.max.hour,
            time.max.minute,
            time.max.second,
        )
        academic_year_start_date = datetime(
            academic_year_end_date.year - 1,
            ACADEMIC_YEAR['start_month'],
            ACADEMIC_YEAR['start_day'],
            time.min.hour,
            time.min.minute,
            time.min.second,
        )

        splitted_interval.append(academic_year_start_date)

    splitted_interval.append(end_date)

    iter_interval = iter(splitted_interval)
    intervals = [*zip(iter_interval, iter_interval)]

    return intervals


def split_interval_by_delta(start_date: datetime, end_date: datetime, days_delta: int) -> List[tuple]:
    """Разбивает учебный год на интервалы по дельте."""
    subinterval_end = start_date + timedelta(days=days_delta)

    splitted_interval = []
    splitted_interval.append(start_date)

    # В данном цикле проверять нужно именно даты чтобы не получилась ситуация что две одинаковые даты
    # попадут в интервалы и различаться будут лишь временем (минимальная дельта 1 день т.е. 24 часа)
    while start_date.date() < subinterval_end.date() < end_date.date():
        splitted_interval.append(subinterval_end)
        subinterval_start = subinterval_end + timedelta(days=1)
        splitted_interval.append(subinterval_start)
        subinterval_end = subinterval_end + timedelta(days=days_delta + 1)

    splitted_interval.append(end_date)

    iter_interval = iter(splitted_interval)
    intervals = [*zip(iter_interval, iter_interval)]

    return intervals


def camel_to_underscore(name: str, upper=False) -> str:
    """Переводит строку из верблюжьего в змеиный регистр.
    По умолчанию строка приводится к нижнему регистру.
    Несколько заглавных букв не будут разъединены, кроме последней.

    Args:
        name: форматируемая строка;
        upper: флаг перевода в верхний регистр;

    Returns:
        Строка, приведенная к змеиному регистру.

    >>> camel_to_underscore('ОдинДваТри')
    'один_два_три'
    >>> camel_to_underscore('Один_два_три')
    'один_два_три'
    >>> camel_to_underscore('Один_Два_Три')
    'один_два_три'
    >>> camel_to_underscore('__ОдинДваТри__')
    '__один_два_три__'
    >>> camel_to_underscore('__ОдинДВАТри__')
    '__один_два_три__'
    """
    output = []
    for i, c in enumerate(name):
        if i > 0:
            pc = name[i - 1]
            if c.isupper() and not pc.isupper() and pc != '_':
                output.append('_')
            elif i > 3 and not c.isupper():
                previous = name[i - 3:i]
                if previous.isalpha() and previous.isupper():
                    output.insert(len(output) - 1, '_')

        output.append(c.upper() if upper else c.lower())

    return ''.join(output)


def build_related_model_graph(
    graph: 'Graph',
    current_model: 'Model',
    find_model: 'Model',
    ignored_model_fields: Optional[Dict[str, Set[str]]] = None,
) -> 'Graph':
    """Построение графа связей от модели до модели.

    Построение графа связей между моделями. Связи строятся только для полей c прямым отношением.
    Обход моделей осуществляется до нахождения нужной или до исчерпания полей со связями (FK, M2M, O2O).

    Args:
        graph: объект Graph
        current_model: объект Model для которой строятся связи
        find_model: объект Model до которой строятся связи
        ignored_model_fields: словарь с наименованием модели (формат app.Model_name)
            и перечисления полей, для которых не нужно строить связь
    """
    if current_model == find_model:
        return graph

    ignored_model_fields = ignored_model_fields or {}
    ignored_current_model_fields = set(ignored_model_fields.get(current_model._meta.label, {}))

    graph.add_vertex(current_model._meta.model_name)

    related_fields = [field for field in current_model._meta.get_fields() if field.is_relation]
    for field in related_fields:
        is_ignore = field.auto_created and not field.concrete and not field.one_to_one
        if is_ignore or field.name in ignored_current_model_fields:
            continue

        related_model = field.related_model

        if related_model._meta.model_name not in graph:
            graph.add_vertex(related_model._meta.model_name)
            graph.add_edge(current_model._meta.model_name, related_model._meta.model_name, field.name)
            graph = build_related_model_graph(graph, related_model, find_model, ignored_model_fields)
        else:
            graph.add_edge(current_model._meta.model_name, related_model._meta.model_name, field.name)

    return graph


def get_data_command_progress_attachment_path(
    instance: Union['EduRdmCollectDataCommandProgress', 'EduRdmExportDataCommandProgress'],
    filename: str
) -> str:
    """Возвращает путь до файла-вложения при формировании команды сборки или выгрузки данных.

    Args:
        instance: объект ExportDataCommandProgress или объект CollectDataCommandProgres
        filename: имя загружаемого файла

    Returns:
        Строковое представление пути
    """
    date_now = datetime.now()

    return os.path.join(
        settings.UPLOADS,
        date_now.strftime('%Y/%m/%d'),
        instance.__class__.__name__.lower(),
        str(instance.type),
        filename
    )


class MakeInterval(Func):
    """Функция обработки даты/времени."""

    template = 'make_interval(%(expressions)s)'

    def __init__(
        self, *,
        years: Union[int, F] = 0,
        months: Union[int, F] = 0,
        weeks: Union[int, F] = 0,
        days: Union[int, F] = 0,
        hours: Union[int, F] = 0,
        minutes: Union[int, F] = 0,
        seconds: Union[float, F] = 0.0,
        output_field=None,
    ) -> None:
        self.years = years
        self.months = months
        self.weeks = weeks
        self.days = days
        self.hours = hours
        self.minutes = minutes
        self.seconds = seconds

        super().__init__(
            self.years,
            self.months,
            self.weeks,
            self.days,
            self.hours,
            self.minutes,
            self.seconds,
            output_field=output_field or DateTimeField(),
        )


def make_passed_datetime_from_today(
    *,
    years: int = 0,
    months: int = 0,
    days: int = 0,
) -> Union[datetime, Expression, F]:
    """Возвращает выражение на лет/месяцев/дней меньше относительно сегодня."""
    now_expr = Now()

    return now_expr - MakeInterval(years=years, months=months, days=days)


TODAY_EXPR: Union[date, Expression, F] = Cast(make_passed_datetime_from_today(), output_field=DateField())

THREE_YEARS_AGO_DATE_EXPR: Union[date, Expression, F] = Cast(
    make_passed_datetime_from_today(years=3),
    output_field=DateField()
)



from typing import (
    NamedTuple,
)


class UploadFile(NamedTuple):
    """Структура файла подэтапа для очереди отправки."""
    attachment_id: int
    attachment_name: str
    attachment_size: int
    operation: str



from abc import (
    abstractmethod,
)
from collections import (
    defaultdict,
)
from typing import (
    Union,
)

from educommon.utils.conversion import (
    int_or_none,
)

from edu_rdm_integration.adapters.caches import (
    WebEduFunctionCacheStorage,
    WebEduRunnerCacheStorage,
)
from edu_rdm_integration.collect_data.base.mixins import (
    ReformatLogsMixin,
)


class BaseCollectingCalculatedExportedDataRunnerCacheStorage(WebEduRunnerCacheStorage):
    """
    Базовый кеш помощников ранеров функций сбора расчетных данных для интеграции с "Региональная витрина данных".
    """


class BaseCollectingCalculatedExportedDataFunctionCacheStorage(ReformatLogsMixin, WebEduFunctionCacheStorage):
    """
    Базовый кеш помощников функций сбора расчетных данных для интеграции с "Региональная витрина данных".
    """

    def __init__(self, raw_logs, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Необработанные логи как есть.
        self.raw_logs = raw_logs

        # Подготовленные логи в виде:
        # {
        #     'person.person': {
        #         48939: [(1, {'surname': 'Иванов', 'snils': '157-283-394 92'...}), (2, {...})],
        #         44281: [(2, {...}), (2, {...}), ...],
        #         12600: [(3, {...})],
        #     },
        #     'schoolchild.schoolchild': {
        #         ...
        #     },
        # }
        self.logs = defaultdict(lambda: defaultdict(list))

    def _ignore_logs(self):
        """Исключение логов из обработки на основании описанных правил."""

    def _prepare_logs(self):
        """Подготовка логов для дальнейшей работы."""
        self._reformat_logs()
        self._ignore_logs()

    @staticmethod
    def _add_id_to_set(ids: set, object_id: Union[int, str, None]) -> None:
        """Добавление значения id во множество ids."""
        object_id = int_or_none(object_id)
        if object_id:
            ids.add(object_id)

    @abstractmethod
    def _collect_product_model_ids(self):
        """Собирает идентификаторы записей моделей продукта с упором на логи."""

    @abstractmethod
    def _prepare_caches(self):
        """Формирование кэшей."""

    def _prepare(self, *args, **kwargs):
        """Запускает формирование кэша помощника Функции."""
        self._prepare_logs()
        self._collect_product_model_ids()
        self._prepare_caches()



from collections import (
    defaultdict,
)
from typing import (
    TYPE_CHECKING,
    Dict,
    Optional,
    Type,
)

import celery
from celery.schedules import (
    crontab,
)
from django.conf import (
    settings,
)
from django.core.cache import (
    cache,
)
from django.utils import (
    timezone,
)

from educommon.async_task.models import (
    AsyncTaskType,
    RunningTask,
)
from educommon.async_task.tasks import (
    PeriodicAsyncTask,
    UniquePeriodicAsyncTask,
)
from educommon.utils.date import (
    get_today_min_datetime,
)
from m3_db_utils.consts import (
    DEFAULT_ORDER_NUMBER,
)

from edu_rdm_integration.collect_and_export_data.models import (
    EduRdmCollectDataCommandProgress,
    EduRdmExportDataCommandProgress,
)
from edu_rdm_integration.collect_data.collect import (
    BaseCollectLatestModelsData,
)
from edu_rdm_integration.collect_data.helpers import (
    set_failed_status_suspended_collecting_data_stages,
)
from edu_rdm_integration.consts import (
    REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA,
    REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA,
    TASK_QUEUE_NAME,
)
from edu_rdm_integration.enums import (
    CommandType,
    FileUploadStatusEnum,
)
from edu_rdm_integration.export_data.export import (
    ExportLatestEntitiesData,
    UploadEntitiesData,
)
from edu_rdm_integration.export_data.helpers import (
    set_failed_status_suspended_exporting_data_stages,
)
from edu_rdm_integration.export_data.queue import (
    RdmRedisSubStageAttachmentQueue,
)
from edu_rdm_integration.helpers import (
    UploadStatusHelper,
    get_collecting_managers_max_period_ended_dates,
    get_exporting_managers_max_period_ended_dates,
    save_command_log_link,
)
from edu_rdm_integration.models import (
    ExportingDataSubStageUploaderClientLog,
    RegionalDataMartEntityEnum,
    TransferredEntity,
    UploadDataCommand,
)
from edu_rdm_integration.storages import (
    RegionalDataMartEntityStorage,
)


if TYPE_CHECKING:
    from datetime import (
        datetime,
    )

    from function_tools.managers import (
        RunnerManager,
    )


class RDMCheckUploadStatus(UniquePeriodicAsyncTask):
    """Периодическая задача для сбора статусов по загрузке файла в витрину."""

    queue = TASK_QUEUE_NAME
    routing_key = TASK_QUEUE_NAME
    description = 'Сбор статусов загрузки данных в витрину "Региональная витрина данных"'
    lock_expire_seconds = settings.RDM_UPLOAD_STATUS_TASK_LOCK_EXPIRE_SECONDS
    task_type = AsyncTaskType.UNKNOWN
    run_every = crontab(
        minute=settings.RDM_UPLOAD_STATUS_TASK_MINUTE,
        hour=settings.RDM_UPLOAD_STATUS_TASK_HOUR,
        day_of_week=settings.RDM_UPLOAD_STATUS_TASK_DAY_OF_WEEK,
    )

    def process(self, *args, **kwargs):
        """Выполнение."""
        super().process(*args, **kwargs)

        # Получаем незавершенные загрузки данных в витрину
        in_progress_uploads = ExportingDataSubStageUploaderClientLog.objects.filter(
            file_upload_status=FileUploadStatusEnum.IN_PROGRESS,
            is_emulation=False,
        ).select_related('attachment')

        UploadStatusHelper(in_progress_uploads, cache).run()


class CheckSuspendedExportedStagePeriodicTask(UniquePeriodicAsyncTask):
    """Периодическая задача поиска зависших этапов/подэтапов экспорта."""

    queue = TASK_QUEUE_NAME
    routing_key = TASK_QUEUE_NAME
    description = 'Поиск зависших этапов/подэтапов экспорта в "Региональная витрина данных"'
    lock_expire_seconds = settings.RDM_CHECK_SUSPEND_TASK_LOCK_EXPIRE_SECONDS
    task_type = AsyncTaskType.SYSTEM
    run_every = crontab(
        minute=settings.RDM_CHECK_SUSPEND_TASK_MINUTE,
        hour=settings.RDM_CHECK_SUSPEND_TASK_HOUR,
        day_of_week=settings.RDM_CHECK_SUSPEND_TASK_DAY_OF_WEEK,
    )

    def process(self, *args, **kwargs):
        """Выполнение задачи."""
        super().process(*args, **kwargs)

        change_status_collecting_result = set_failed_status_suspended_collecting_data_stages()
        change_status_exporting_result = set_failed_status_suspended_exporting_data_stages()

        task_result = {
            'Прервано сборок': (
                f'Этапов {change_status_collecting_result["change_stage_count"]}'
                f' и подэтапов {change_status_collecting_result["change_sub_stage_count"]}'
            ),
            'Прервано выгрузок': (
                f'Этапов {change_status_exporting_result["change_stage_count"]}'
                f' и подэтапов {change_status_exporting_result["change_sub_stage_count"]}'
            ),
        }

        self.set_progress(
            values=task_result
        )


class TransferLatestEntitiesDataPeriodicTask(UniquePeriodicAsyncTask):
    """Периодическая задача сбора и выгрузки данных."""

    queue = TASK_QUEUE_NAME
    routing_key = TASK_QUEUE_NAME
    description = 'Периодическая задача сбора и экспорта данных РВД'
    lock_expire_seconds = settings.RDM_TRANSFER_TASK_LOCK_EXPIRE_SECONDS
    task_type = AsyncTaskType.UNKNOWN
    run_every = crontab(
        minute=settings.RDM_TRANSFER_TASK_MINUTE,
        hour=settings.RDM_TRANSFER_TASK_HOUR,
        day_of_week=settings.RDM_TRANSFER_TASK_DAY_OF_WEEK,
    )

    def __init__(self) -> None:
        super().__init__()

        self._collecting_data_managers: Dict[str, Type['RunnerManager']] = {}
        self._collecting_data_manager_to_logs_period_end: Dict[str, 'datetime'] = {}

        self._exporting_data_managers: Dict[str, Type['RunnerManager']] = {}
        self._exporting_data_manager_to_period_end: Dict[str, 'datetime'] = {}

        self._transferred_entities = []
        self._entites_models_map = defaultdict(list)

    def process(self, *args, **kwargs):
        """Выполняет задачу."""
        super().process(*args, **kwargs)

        self._collect_transferred_entities()
        self._collect_managers()
        self._calculate_collecting_managers_logs_period_ended_at()
        self._calculate_exporting_managers_ended_at()

        task_id = RunningTask.objects.filter(
            pk=self.request.id,
        ).values_list('pk', flat=True).first()

        collected_entity_models = set()

        for entity_enum, export_enabled in sorted(
            self._transferred_entities, key=lambda entity: entity[0].order_number
        ):
            entity_models = self._entites_models_map.get(entity_enum.key, ())
            for model_enum_value in entity_models:
                if model_enum_value.key not in collected_entity_models:
                    self._run_collect_model_data(model_enum_value.key, task_id)
                    collected_entity_models.add(model_enum_value.key)

            try:
                if export_enabled:
                    self._run_export_entity_data(entity_enum.key, task_id)
            except Exception:
                continue

    def _run_collect_model_data(self, model: str, task_id: str) -> None:
        """Запускает сбор данных модели РВД."""
        command = self._create_collect_command(model, task_id)
        collect_model_data = self._prepare_collect_model_data_class(command)
        collect_model_data.collect()

        command.refresh_from_db(fields=['stage_id'])
        save_command_log_link(command, settings.RDM_COLLECT_LOG_DIR)

    def _run_export_entity_data(self, entity: str, task_id: str) -> None:
        """Запускает экспорт данных сущности РВД."""
        command = self._create_export_command(entity, task_id)
        if command:
            export_entity_data = self._prepare_export_entity_data_class(command)
            export_entity_data.export()

            command.refresh_from_db(fields=['stage_id'])
            save_command_log_link(command, settings.RDM_EXPORT_LOG_DIR)

    def _collect_transferred_entities(self) -> None:
        """Собирает сущности РВД, по которым будет произведен сбор и экспорт данных."""
        self._transferred_entities = [
            (RegionalDataMartEntityEnum.get_model_enum_value(key=entity), export_enabled)
            for entity, export_enabled in TransferredEntity.objects.values_list('entity', 'export_enabled')
        ]

        # Собираем словарь по сущностям с моделями для сборки
        for entity, _ in self._transferred_entities:
            self._entites_models_map[entity.key].extend(
                (model_enum for model_enum in (entity.main_model_enum, *entity.additional_model_enums)
                 if model_enum.order_number != DEFAULT_ORDER_NUMBER)
            )

    def _collect_managers(self) -> None:
        """Собирает менеджеры Функций для сбора и выгрузки данных."""
        entity_storage = RegionalDataMartEntityStorage()
        entity_storage.prepare()

        collecting_models_data_managers_map = entity_storage.prepare_entities_manager_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA},
        )
        exporting_entities_data_managers_map = entity_storage.prepare_entities_manager_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA},
        )

        for entity_key, entity_models in self._entites_models_map.items():
            for entity_model in entity_models:
                collect_manager_class = collecting_models_data_managers_map.get(entity_model.key)
                if collect_manager_class:
                    self._collecting_data_managers[entity_model.key] = collect_manager_class

            export_manager_class = exporting_entities_data_managers_map.get(entity_key)
            if export_manager_class:
                self._exporting_data_managers[entity_key] = export_manager_class

    def _calculate_collecting_managers_logs_period_ended_at(self) -> None:
        """Определяет дату последнего успешного этапа сбора у менеджеров Функций сбора."""
        self._collecting_data_manager_to_logs_period_end = get_collecting_managers_max_period_ended_dates(
            self._collecting_data_managers.values()
        )

    def _calculate_exporting_managers_ended_at(self) -> None:
        """Определяет дату последнего успешного подэтапа экспорта у менеджеров Функций экспорта."""
        self._exporting_data_manager_to_period_end = get_exporting_managers_max_period_ended_dates(
            self._exporting_data_managers.values()
        )

    def _create_collect_command(self, model: str, task_id: str) -> EduRdmCollectDataCommandProgress:
        """Создает команду сбора данных моделей РВД."""
        manager = self._collecting_data_managers[model]
        manager_last_collected = (
            self._collecting_data_manager_to_logs_period_end.get(manager.uuid)
            or get_today_min_datetime()
        )

        period_started_at = manager_last_collected
        period_ended_at = timezone.now()

        return EduRdmCollectDataCommandProgress.objects.create(
            model_id=model,
            logs_period_started_at=period_started_at,
            logs_period_ended_at=period_ended_at,
            task_id=task_id,
            type=CommandType.AUTO,
        )

    def _create_export_command(self, entity: str, task_id: str) -> Optional[EduRdmExportDataCommandProgress]:
        """Создает команду экспорта данных сущностей РВД."""
        manager = self._exporting_data_managers[entity]
        manager_last_exported = self._exporting_data_manager_to_period_end.get(manager.uuid)

        if manager_last_exported:
            period_started_at = manager_last_exported
            period_ended_at = timezone.now()

            return EduRdmExportDataCommandProgress.objects.create(
                entity_id=entity,
                period_started_at=period_started_at,
                period_ended_at=period_ended_at,
                task_id=task_id,
                type=CommandType.AUTO,
            )

    def _prepare_collect_model_data_class(
        self,
        command: EduRdmCollectDataCommandProgress
    ) -> BaseCollectLatestModelsData:
        """Подготавливает объект класса сбора данных моделей РВД."""
        return BaseCollectLatestModelsData(
            models=[command.model_id],
            logs_period_started_at=command.logs_period_started_at,
            logs_period_ended_at=command.logs_period_ended_at,
            command_id=command.id,
            use_times_limit=True,
        )

    def _prepare_export_entity_data_class(self, command: EduRdmExportDataCommandProgress) -> ExportLatestEntitiesData:
        """Подготавливает объект класса экспорта данных сущностей РВД.

        При экспорте данных передаем параметр task_id для обновления поля "Описание"
        наименованиями выгруженных сущностей.
        """
        return ExportLatestEntitiesData(
            entities=[command.entity_id],
            period_started_at=command.period_started_at,
            period_ended_at=command.period_ended_at,
            command_id=command.id,
            task_id=self.request.id,
        )


class UploadDataAsyncTask(UniquePeriodicAsyncTask):
    """Формирование очереди файлов и их отправка."""

    queue = TASK_QUEUE_NAME
    routing_key = TASK_QUEUE_NAME
    description = 'Отправка данных в витрину "Региональная витрина данных"'
    lock_expire_seconds = settings.RDM_UPLOAD_DATA_TASK_LOCK_EXPIRE_SECONDS
    task_type = AsyncTaskType.SYSTEM
    run_every = crontab(
        minute=settings.RDM_UPLOAD_DATA_TASK_MINUTE,
        hour=settings.RDM_UPLOAD_DATA_TASK_HOUR,
        day_of_week=settings.RDM_UPLOAD_DATA_TASK_DAY_OF_WEEK,
    )

    def process(self, *args, **kwargs):
        """Выполнение."""
        super().process(*args, **kwargs)

        entity_ids_for_export = list(
            TransferredEntity.objects.filter(export_enabled=True).values_list('entity_id', flat=True)
        )
        queue = RdmRedisSubStageAttachmentQueue()
        task_id = RunningTask.objects.filter(
            pk=self.request.id,
        ).values_list('pk', flat=True).first()

        if task_id:
            for entity_id in entity_ids_for_export:
                upload_data_command = UploadDataCommand.objects.create(
                    entity_id=entity_id,
                    task_id=task_id,
                    type=CommandType.AUTO,
                )
                if upload_data_command:
                    upload_data = UploadEntitiesData(
                        entities=[upload_data_command.entity_id],
                        data_cache=cache,
                        queue=queue,
                        task_id=task_id,
                        command_id=upload_data_command.id,
                    )
                    upload_data.upload_data()
                    save_command_log_link(upload_data_command, settings.RDM_UPLOAD_LOG_DIR)


celery_app = celery.app.app_or_default()
celery_app.register_task(RDMCheckUploadStatus)
celery_app.register_task(CheckSuspendedExportedStagePeriodicTask)
celery_app.register_task(TransferLatestEntitiesDataPeriodicTask)
celery_app.register_task(UploadDataAsyncTask)
